import fractions


class Matrice:

    # Constructeur de la classe
    def __init__(self, n, m):
        """Une matrice s'initialise en donnant sa dimension : A = Matrice(2, 3).
        Tous les éléménts de la matrice sont initialisés à 0.
        """
        assert n > 0, "Le nombre de lignes de la matrice doit être positif."
        assert m > 0, "Le nombre de colonnes de la matrice doit être positif."

        self._nb_lignes = n
        self._nb_colonnes = m
        self._elements = []
        for i in range(self._nb_lignes):
            ligne = []
            for j in range(self._nb_colonnes):
                ligne.append(0)
            self._elements.append(ligne)

    # Accesseur pour le nombre de lignes
    def GetnNb_lignes(self):
        """ A.getnNb_lignes() retourne le nombre de lignes de la Matrice A."""
        return self._nb_lignes

    # Accesseur pour le nombre de colonnes
    def GetNb_colonnes(self):
        """ A.getNb_colonnes() retourne le nombre de colonnes de la Matrice A."""
        return self._nb_colonnes

    # Accesseur pour la valeur du terme d'indice ij.
    def GetIJeme(self, i, j):
        """ A.GetIJeme(i, j) retourne la valeur du terme d'indice ij de la Matrice A.
        L'indice des lignes et des colonnes commence à 1."""
        assert 0 <= i-1 < self._nb_lignes, "L'indice des lignes n'est pas valide."
        assert 0 <= j-1 < self._nb_colonnes, "L'indice des colonnes n'est pas valide."

        return self._elements[i-1][j-1]

    # Accesseur pour la ligne d'indice i.
    def GetIeme_ligne(self, i):
        """ A.GetIeme_ligne(i) retourne la ieme ligne (list) de la Matrice A.
        L'indice des lignes commence à 1."""
        assert 0 <= i-1 < self._nb_lignes, "L'indice de la ligne n'est pas valide."

        return self._elements[i-1]

    # Accesseur pour la colonne d'indice j.
    def GetJeme_colonne(self, j):
        """ A.GetJeme_colonne(j) retourne la jeme colonne (list) de la Matrice A.
        L'indice des colonnes commence à 1."""
        assert 0 <= j-1 < self._nb_colonnes, "L'indice de la colonne n'est pas valide."
        colonne = []

        for i in range(self._nb_lignes):
            colonne.append(self._elements[i][j-1])
        return colonne

    # Mutateur pour la valeur du terme d'indice ij.
    def SetIJeme(self, i, j, e):
        """ A.SetIJeme(i, j, e) modifie le terme d'indice ij de la Matrice A
        par la valeur (e).
        L'indice des lignes et des colonnes commence à 1."""
        assert 0 <= i-1 < self._nb_lignes, "l'indice des lignes n'est pas valide."
        assert 0 <= j-1 < self._nb_colonnes, "l'indice des colonnes n'est pas valide."
        assert isinstance(e, int) or isinstance(e, float) \
            or isinstance(e, fractions.Fraction), "l'élément n'est pas un nombre."

        self._elements[i-1][j-1] = e

    # Mutateur pour la ligne d'indice i.
    def SetIeme_ligne(self, valeurs, i):
        """ A.SetIeme_ligne(valeurs, i) modifie la ligne d'indice i de la Matrice A
        par valeurs (list). L'indice des lignes commence à 1."""
        assert 0 <= i-1 < self._nb_lignes, "L'indice de la ligne n'est pas valide."
        assert isinstance(valeurs, list), "Les valeurs doivent être formatées en liste."
        assert len(valeurs) == self._nb_colonnes, "Le nombre de valeurs est incohérent."

        for j in range(self._nb_colonnes):
            self.SetIJeme(i, j+1, valeurs[j])

    # Mutateur pour la colonne d'indice j.
    def SetJeme_colonne(self, valeurs, j):
        """ A.SetJeme_colonne(valeurs, j) modifie la colonne d'indice j de la Matrice A
        par valeurs (list). L'indice des colonnes commence à 1."""
        assert 0 <= j-1 < self._nb_colonnes, "L'indice de la colonne n'est pas valide."
        assert isinstance(valeurs, list), "Les valeurs doivent être formatées en liste."
        assert len(valeurs) == self._nb_lignes, "Le nombre de valeurs est incohérent."

        for i in range(self._nb_lignes):
            self.SetIJeme(i+1, j, valeurs[i])

    # Méthodes privées de la classe
    # Somme de deux Matrices
    def __add__(self, B):
        """ Définition du comportement du symbole "+" entre deux Matrices."""
        assert isinstance(B, Matrice), \
            "Vos deux opérandes doivent être des instances de la classe Matrice."
        assert self._nb_lignes == B._nb_lignes and self._nb_colonnes == B._nb_colonnes, \
            "Il y a une incohérence de dimensions."

        Somme = Matrice(self._nb_lignes, self._nb_colonnes)
        for i in range(self._nb_lignes):
            for j in range(self._nb_colonnes):
                Somme.SetIJeme(i+1, j+1, self.GetIJeme(i+1, j+1) + B.GetIJeme(i+1, j+1))
        return Somme

    # Produit de deux matrices
    # Fonction auxiliaire pour le produit
    def __prod_ligne_colonne__(L, C):
        """Multiplication d'une ligne (list) par une colonne (list)."""
        resultat = 0

        for i in range(len(L)):
            resultat += L[i]*C[i]
        return resultat

    def __mul__(self, B):
        """Définition du comportement du symbole "*" entre deux Matrices."""
        assert isinstance(B, Matrice) or isinstance(B, int) or isinstance(B, float) or isinstance(B, fractions.Fraction), \
            "Il y a un problème dans la nature de vos opérandes."
        if isinstance(B, Matrice):
            assert self._nb_colonnes == B._nb_lignes, "Il y a une incohérence de dimensions."

            Produit = Matrice(self._nb_lignes, B._nb_colonnes)
            for i in range(self._nb_lignes):
                L = self.GetIeme_ligne(i+1)
                for j in range(B._nb_colonnes):
                    C = B.GetJeme_colonne(j+1)
                    Produit.SetIJeme(i+1, j+1, Matrice.__prod_ligne_colonne__(L, C))
        else:
            Produit = Matrice(self._nb_lignes, self._nb_colonnes)
            for i in range(self._nb_lignes):
                for j in range(self._nb_colonnes):
                    Produit.SetIJeme(i+1, j+1, B*self.GetIJeme(i+1, j+1))
        return Produit

    # A*A...*A
    def __pow__(self, n):
        assert isinstance(n, int), "L'exposant doit être un entier naturel."
        assert self._nb_lignes == self._nb_colonnes, "La Matrice doit être carrée."
        if n == 0:
            M = Matrice(self._nb_lignes, self._nb_colonnes)
            for i in range(M._nb_lignes):
                M.SetIJeme(i+1, i+1, 1)
        else:
            M = self
            for i in range(n-1):
                M = M * self
        return M

    # Test de la présence d'un élément dans une Matrice
    def __contains__(self, element):
        """Définition du comportement du mot-clé "in" entre un élément et une Matrice.
        Pour la recherche d'une fraction : Fraction(1, 3) in A."""
        for i in range(self._nb_lignes):
            if element in self._elements[i]:
                return True
        return False

    # Test d'égalité de deux matrices
    def __eq__(self, mat):
        """Définition du comportement du test "==" entre deux Matrices."""
        return self._nb_lignes == mat._nb_lignes and self._nb_colonnes == mat._nb_colonnes and self._elements == mat._elements

    # Représentation de l'état de l'instance
    def __str__(self):
        """Comportement de l'instruction "print(A)."""
        return self._elements.__str__()

    # Schéma de construction de l'instance
    def __repr__(self):
        """Comportement de l'instruction "A" dans un interpréteur."""
        return f"Matrice({str(self._nb_lignes)}, {str(self._nb_colonnes)})"

    # Affichage de l'instance sous une forme plus matricielle
    # Fonction auxiliaire 1
    def __elements__max__(self):
        """Retourne la taille du plus grand élement de chaque colonne d'une Matrice."""
        L = []
        for j in range(self._nb_colonnes):
            tailles = []
            for i in range(self._nb_lignes):
                tailles.append(len(str(self.GetIJeme(i+1, j+1))))
            L.append(max(tailles))
        return L

    # Fonction auxiliaire 2
    def __formatage_ligne__(self, i):
        """Retourne un affichage (str) formaté de la ligne d'indice i d'une matrice.
        L'indice des lignes commence à 1."""
        L = self.__elements__max__()
        affichage = "|"
        for j in range(len(self._elements[i-1])):
            elem = self._elements[i-1][j]
            nb_espaces = L[j] - len(str(elem))
            elem_formatage = ""
            for a in range(nb_espaces):
                elem_formatage += " "
            elem_formatage += f"{elem}"
            affichage += f" {elem_formatage} "
        affichage += "|"
        return affichage

    # Méthodes publiques de la classe
    # Recherche des indices d'un élément existant dans une Matrice
    def recherche_indices_element(self, element):
        """Retourne sous forme de liste les indices tuples d'un élémént
        existant dans une Matrice."""
        assert self.__contains__(element), f"{element} n'est pas un élément de la matrice."

        L = []
        for i in range(self._nb_lignes):
            for j in range(self._nb_colonnes):
                if self.GetIJeme(i+1, j+1) == element:
                    L.append((i+1, j+1))
        return L

    # Affichage de l'instance sous une forme plus matricielle
    # Méthode publique
    def affichage_matricielle(self):
        for i in range(1, self._nb_lignes + 1):
            print(self.__formatage_ligne__(i))

    # Affichage en valeur approchée des élément d'une ligne
    def affichage_ligne_approximation(self, i):
        """Affichage des élements d'une ligne d'une Matrice
        avec les éventuelles fractions en écriture décimale
        au moins approchée."""
        affichage = "| "
        for elem in self.GetIeme_ligne(i):
            if isinstance(elem, fractions.Fraction):
                affichage += f"{str(float(elem))} "
            else:
                affichage += f"{str(elem)} "
        affichage += "|"
        print(affichage)

    # Matrice identité
    def matrice_identite(self):
        """A.matrice_identite() fait de A une matrice identité."""
        assert (self._nb_colonnes == self._nb_lignes), "La Matrice doit être carrée."
        for i in range(1, self._nb_lignes + 1):
            self.SetIJeme(i, i, 1)
